﻿using System;
using System.Linq.Expressions;

namespace Draki.Interfaces
{
    public interface IActionSyntaxProvider : ISyntaxProvider
    {
        // native only
        /// <summary>
        /// Click a specified element.
        /// </summary>
        /// <param name="element"><see cref="IElement"/> factory function.</param>
        IActionSyntaxProvider Click(ElementProxy element);

        /// <summary>
        /// Click a specified button within an alert.
        /// </summary>
        /// <param name="accessor"></param>
        IActionSyntaxProvider Click(Alert accessor);

        /// <summary>
        /// DoubleClick a specified element.
        /// </summary>
        /// <param name="element"><see cref="IElement"/> factory function.</param>
        IActionSyntaxProvider DoubleClick(ElementProxy element);

        /// <summary>
        /// RightClick a specified element.
        /// </summary>
        /// <param name="element"><see cref="IElement"/> factory function.</param>
        IActionSyntaxProvider RightClick(ElementProxy element);

        /// <summary>
        /// Begin a Drag/Drop operation starting with the specified element.
        /// </summary>
        /// <param name="element"><see cref="IElement"/> factory function.</param>
        /// <returns><c>DragDropSyntaxProvider</c></returns>
        ActionSyntaxProvider.DragDropSyntaxProvider Drag(ElementProxy element);

        /// <summary>
        /// Find an element matching <paramref name="selector"/>.
        /// </summary>
        /// <param name="selector">Sizzle DOM selector.</param>
        /// <returns>IElement factory function for lazy access to elements.</returns>
        ElementProxy Find(string selector);

        /// <summary>
        /// Find a set of elements matching <paramref name="selector"/>.
        /// </summary>
        /// <param name="selector"></param>
        /// <returns>IElement factory function for lazy access to a collection of elements.</returns>
        ElementProxy FindMultiple(string selector);

        /// <summary>
        /// Sets the focus to a specific element.
        /// </summary>
        /// <param name="element">IElement factory function.</param>
        IActionSyntaxProvider Focus(ElementProxy element);

        /// <summary>
        /// Causes the mouse to hover over a specified element.
        /// </summary>
        /// <param name="element">IElement factory function.</param>
        IActionSyntaxProvider Hover(ElementProxy element);

        /// <summary>
        /// Wait until the provided <paramref name="conditionAction">action</paramref> succeeds. Intended for use with I.Expect.* methods.
        /// </summary>
        /// <param name="conditionAction">Action to be repeated until it succeeds or exceeds the timeout. <see cref="Settings.DefaultWaitUntilTimeout"/> determines the timeout.</param>
        IActionSyntaxProvider WaitUntil(Expression<Action> conditionAction);

        /// <summary>
        /// Wait until the provided <paramref name="conditionAction">action</paramref> succeeds. Intended for use with I.Expect.* methods.
        /// </summary>
        /// <param name="conditionAction">Action to be repeated until it succeeds or exceeds the timeout.</param>
        /// <param name="timeout">Timeout for this specific action.</param>
        IActionSyntaxProvider WaitUntil(Expression<Action> conditionAction, TimeSpan timeout);

        /// <summary>
        /// Wait until the provided <paramref name="conditionAction">action</paramref> succeeds. Intended for use with I.Expect.* methods.
        /// </summary>
        /// <param name="conditionAction">Action to be repeated until it succeeds or exceeds the timeout.</param>
        /// <param name="seconds">Timeout in seconds for this specific action.</param>
        IActionSyntaxProvider WaitUntil(Expression<Action> conditionAction, int seconds);

        /// <summary>
        /// Wait until the provided <paramref name="conditionFunc">function</paramref> returns <c>true</c>.
        /// </summary>
        /// <param name="conditionFunc">Function to be repeated until it returns true or exceeds the timeout. <see cref="Settings.DefaultWaitUntilTimeout"/> determines the timeout.</param>
        IActionSyntaxProvider WaitUntil(Expression<Func<bool>> conditionFunc);

        /// <summary>
        /// Wait until the provided <paramref name="conditionFunc">function</paramref> returns <c>true</c>.
        /// </summary>
        /// <param name="conditionFunc">Function to be repeated until it returns true or exceeds the timeout.</param>
        /// <param name="seconds">Timeout in seconds for this specific action.</param>
        IActionSyntaxProvider WaitUntil(Expression<Func<bool>> conditionFunc, int seconds);

        /// <summary>
        /// Wait until the provided <paramref name="conditionFunc">function</paramref> returns <c>true</c>.
        /// </summary>
        /// <param name="conditionFunc">Function to be repeated until it returns true or exceeds the timeout.</param>
        /// <param name="timeout">Timeout for this specific action.</param>
        IActionSyntaxProvider WaitUntil(Expression<Func<bool>> conditionFunc, TimeSpan timeout);

        //// remote commands
        ///// <summary>
        ///// Click at the specified coordinates.
        ///// </summary>
        ///// <param name="x">X-coordinate.</param>
        ///// <param name="y">Y-coordinate</param>
        //IActionSyntaxProvider Click(int x, int y);
        // #ADH Not currently working with the latest Chrome65, please use Click(element, x,y) instead.

        /// <summary>
        /// Click the element matching <paramref name="selector"/>.
        /// </summary>
        /// <param name="selector">Sizzle selector.</param>
        IActionSyntaxProvider Click(string selector);

        /// <summary>
        /// DoubleClick the element matching <paramref name="selector"/>.
        /// </summary>
        /// <param name="selector">Sizzle selector.</param>
        IActionSyntaxProvider DoubleClick(string selector);

        /// <summary>
        /// RightClick the element matching <paramref name="selector"/>.
        /// </summary>
        /// <param name="selector">Sizzle selector.</param>
        IActionSyntaxProvider RightClick(string selector);

        /// <summary>
        /// Begin a Drag/Drop operation starting with the element matching <paramref name="selector"/>.
        /// </summary>
        /// <param name="selector">Sizzle selector.</param>
        /// <returns><c>DragDropSyntaxProvider</c></returns>
        ActionSyntaxProvider.DragDropSyntaxProvider Drag(string selector);

        /// <summary>
        /// Enter a number or other object value into a valid input or textarea. Syntactical candy to avoid having to call .ToString() on integers in tests.
        /// </summary>
        /// <param name="nonString">Value to enter into input or textarea.</param>
        /// <returns><c>TextEntrySyntaxProvider</c></returns>
        ActionSyntaxProvider.TextEntrySyntaxProvider Enter(dynamic nonString);

        /// <summary>
        /// Enter text into a valid input or textarea.
        /// </summary>
        /// <param name="text">Text to enter into input or textarea.</param>
        /// <returns><c>TextEntrySyntaxProvider</c></returns>
        ActionSyntaxProvider.TextEntrySyntaxProvider Enter(string text);

        /// <summary>
        /// Append a number or other object value into a valid input or textarea. Syntactical candy to avoid having to call .ToString() on integers in tests.
        /// </summary>
        /// <param name="nonString">Value to enter into input or textarea.</param>
        /// <returns><c>TextEntrySyntaxProvider</c></returns>
        ActionSyntaxProvider.TextAppendSyntaxProvider Append(dynamic nonString);

        /// <summary>
        /// Append text into a valid input or textarea.
        /// </summary>
        /// <param name="text">Text to enter into input or textarea.</param>
        /// <returns><c>TextEntrySyntaxProvider</c></returns>
        ActionSyntaxProvider.TextAppendSyntaxProvider Append(string text);

        /// <summary>
        /// Sets the focus to element matching <paramref name="selector"/>.
        /// </summary>
        /// <param name="selector">Sizzle selector.</param>
        IActionSyntaxProvider Focus(string selector);

        /// <summary>
        /// Causes the mouse to hover over element matching <paramref name="selector"/>.
        /// </summary>
        /// <param name="selector"></param>
        IActionSyntaxProvider Hover(string selector);

        /// <summary>
        /// Scrolls the viewport to the element matching <paramref name="selector"/>. Alias for Hover(string).
        /// </summary>
        /// <param name="selector"></param>
        IActionSyntaxProvider Scroll(string selector);

        /// <summary>
        /// Scrolls the viewport to the specified <paramref name="element"/>. Alias for Hover(ElementProxy).
        /// </summary>
        /// <param name="element">IElement factory function.</param>
        IActionSyntaxProvider Scroll(ElementProxy element);

        /// <summary>
        /// Open a web browser and navigate to specified URL
        /// </summary>
        /// <param name="url">Fully-qualified URL. Example: <c>http://google.com/</c></param>
        IActionSyntaxProvider Open(string url);

        /// <summary>
        /// Open a web browser and navigate to the specified URI
        /// </summary>
        /// <param name="uri">Absolute URI. Example: <c>new Uri("http://www.google.com/");</c></param>
        IActionSyntaxProvider Open(Uri uri);

        /// <summary>
        /// Triggers special Selenium keypresses.
        /// </summary>
        /// <param name="keys">SendKeys values. Example: <c>{TAB}</c></param>
        /// <example>Supported keys {ENTER},{TAB}</example>
        IActionSyntaxProvider Press(params Key[] keys);

        /// <summary>
        /// Manipulates a <c>&lt;select /></c> DOM element by selecting options with matching <paramref name="values"/> using the specified <paramref name="mode"/>.
        /// </summary>
        /// <param name="mode">Mode of interaction with the <c>&lt;select /></c>; by Text or Value.</param>
        /// <param name="values">Options to be selected.</param>
        /// <returns><c>SelectSyntaxProvider</c></returns>
        ActionSyntaxProvider.SelectSyntaxProvider Select(Option mode, params string[] values);

        /// <summary>
        /// Manipulates a <c>&lt;select /></c> DOM element by selecting an option with matching <paramref name="value"/> using the specified <paramref name="mode"/>.
        /// </summary>
        /// <param name="mode">Mode of interaction with the <c>&lt;select /></c>; by Text or Value.</param>
        /// <param name="value">Option to be selected.</param>
        /// <returns><c>SelectSyntaxProvider</c></returns>
        ActionSyntaxProvider.SelectSyntaxProvider Select(Option mode, string value);

        /// <summary>
        /// Manipulates a <c>&lt;select /></c> DOM element by selecting options at the specified <paramref name="indices"/>.
        /// </summary>
        /// <param name="indices">Options to be selected by Index.</param>
        /// <returns><c>SelectSyntaxProvider</c></returns>
        ActionSyntaxProvider.SelectSyntaxProvider Select(params int[] indices);

        /// <summary>
        /// Manipulates a <c>&lt;select /></c> DOM element by selecting options with matching <paramref name="text"/>.
        /// </summary>
        /// <param name="values">Options to be selected by Text.</param>
        /// <returns><c>SelectSyntaxProvider</c></returns>
        ActionSyntaxProvider.SelectSyntaxProvider Select(params string[] text);

        /// <summary>
        /// Manipulates a <c>&lt;select /></c> DOM element by selecting an option at the specified <paramref name="index"/>.
        /// </summary>
        /// <param name="index">Option to be selected by Index.</param>
        /// <returns><c>SelectSyntaxProvider</c></returns>
        ActionSyntaxProvider.SelectSyntaxProvider Select(int index);

        /// <summary>
        /// Manipulates a <c>&lt;select /></c> DOM element by selecting an option with matching <paramref name="text"/>.
        /// </summary>
        /// <param name="text">Option to be selected by Text.</param>
        /// <returns><c>SelectSyntaxProvider</c></returns>
        ActionSyntaxProvider.SelectSyntaxProvider Select(string text);

        /// <summary>
        /// Takes a screenshot of the active web browser window.
        /// </summary>
        /// <param name="screenshotName">Filename to save the screenshot with.</param>
        IActionSyntaxProvider TakeScreenshot(string screenshotName);

        /// <summary>
        /// Triggers Win32 events per character in the provided string. Not for use with WinForms SendKeys commands, only simple characters. Useful
        /// for entering text into applications that can gain focus but do not have proper DOM representation for use with <see cref="Enter"/>.
        /// </summary>
        /// <param name="text">String to be sent, one character at a time.</param>
        IActionSyntaxProvider Type(string text);

        /// <summary>
        /// Waits the duration of the WaitTimeout as specified in settings.
        /// </summary>
        /// <returns></returns>
        IActionSyntaxProvider Wait();

        /// <summary>
        /// Waits a determined period of time.
        /// </summary>
        /// <param name="seconds">Seconds to wait.</param>
        IActionSyntaxProvider Wait(int seconds);

        /// <summary>
        /// Waits a determined period of time.
        /// </summary>
        /// <param name="timeSpan">TimeSpan to wait.</param>
        IActionSyntaxProvider Wait(TimeSpan timeSpan);

        /// <summary>
        /// Expects - Fluent's passive expect functionality. Defaults to Assert mode (fail on exception) for backwards compatibility. This will change
        /// in a future release. Can be set to passively expect by setting Draki.Settings.ExpectIsAssert = false
        /// </summary>
        AssertSyntaxProvider Expect { get; }

        /// <summary>
        /// Asserts - Fluent's assertion functionality.
        /// </summary>
        AssertSyntaxProvider Assert { get; }

        /// <summary>
        /// Switch to another window or frame.
        /// </summary>
        ActionSyntaxProvider.SwitchSyntaxProvider Switch { get; }

        /// <summary>
        /// return the current page Title
        /// </summary>
        /// <returns></returns>
        string GetTitle();

        /// <summary>
        /// Quick check that a cssSelector exists. You must have first confirmed the page has fully loaded before using this by calling any of the non quick methods.
        /// </summary>
        /// <param name="cssSelector"></param>
        /// <returns>true if the element exists</returns>
        bool QuickExists(string cssSelector);
    }
}
